/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"

/*
 * TODO
 * - support accept filters in FreeBSD
 * - configurable socket buffer sizes (where to put config?)
 * - configurable listen() backlog (where to put config?)
 */
#ifdef TCP_DEFER_ACCEPT
static int set_accept_filter(int fd, const char *accept_filter)
{
	/*
	 * Disabled for now, weird kernel bug with insane/unrealistic
	 * concurrency that causes writev() responses to not go
	 * anywhere for several seconds.
	 *
	 * TODO: investigate and generate a test case for
	 * netdev@vger.kernel.org.
	 */

	return 0;
}
#else /* !TCP_DEFER_ACCEPT */
static int set_accept_filter(int fd, const char *accept_filter)
{
	/* TODO: FreeBSD accept filter */
	return 0;
}
#endif /* !TCP_DEFER_ACCEPT */

static int set_tcp_opts(int fd, const char *accept_filter)
{
	int val;
	socklen_t len = sizeof(int);
	int rc;

	val = 1;
	rc = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &val, len);
	if (rc < 0) return rc;

	val = 1;
	rc = setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &val, len);
	if (rc < 0) return rc;

	val = 1;
	rc = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &val, len);
	if (rc < 0) return rc;

	if (accept_filter) {
		rc = set_accept_filter(fd, accept_filter);
		if (rc < 0) return rc;
	}

	return rc;
}

int mog_bind_listen(struct addrinfo *r, const char *accept_filter)
{
	int fd = -1;

	for (; r; r = r->ai_next) {
		fd = socket(r->ai_family, r->ai_socktype, r->ai_protocol);
		if (fd < 0)
			continue;

		/*
		 * We'll need to unset FD_CLOEXEC in the child for upgrades
		 * Leave FD_CLOEXEC set because we fork+exec iostat(1)
		 * frequently.  We can't guarantee SOCK_CLOEXEC works
		 * everywhere yet (in 2012).
		 */
		if (mog_set_cloexec(fd, true) == 0 &&
		    set_tcp_opts(fd, accept_filter) == 0 &&
		    bind(fd, r->ai_addr, r->ai_addrlen) == 0 &&
		    listen(fd, 1024) == 0)
			break;

		PRESERVE_ERRNO( close(fd) );
		fd = -1;
	}

	return fd;
}
